import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from scipy.stats import iqr
from IPython.display import display
import os

from sklearn.metrics import accuracy_score, confusion_matrix, f1_score, precision_score, recall_score, roc_auc_score
from sklearn.metrics import classification_report

from wordcloud import WordCloud
from PIL import Image


# FUNCION 1
def visualizeME_palettes_or_colors(selection = 'palettes', quantity_colors= 8):
    '''
    Function for consulting the colors and palettes available of the Seaborn library
    ### Parameters (2):
        * selection: `str` by default gives 'palettes', but you can choose 'colors'
        * quantity_colors: `int` by default it returns 8 colors per palette, you can change number of colors you will need. And if you want to see colors, it is not neccesary this parameter.
    ### Return (1):
        * plt.show(): available palettes/ colors with their respective names
    '''
    colors = pd.read_csv('data/viz/seaborn_color_list.csv')

    if selection == 'palettes':
        grid = np.vstack((np.linspace(0, 1, quantity_colors), np.linspace(0, 1, quantity_colors)))
        options_colors = colors['PALETTE_COLORS'].dropna().sort_values(key=lambda x: x.str.lower())
        col = 4                             
        pos = 1 
        row = int(len(options_colors)/col)+1 
        plt.figure(figsize=(col*4,row))

        for i in options_colors:
            if '_r' in i:
                pass
            else:
                plt.subplot(row, col, pos)
                plt.imshow(grid, cmap = i, aspect='auto')
                plt.axis('off')
                plt.title(i, loc = 'center', fontsize = 20)
                pos = pos + 1
    
        print('If you want any palette reversed, just add "_r" at the end of the palette name')

    elif selection == 'colors':
        just_colors = sorted(colors['CSS4_COLORS'].dropna().sort_values(key=lambda x: x.str.lower()))
        col = 4                             
        pos = 1 
        row = int(len(just_colors)/col)+1 
        plt.figure(figsize=(col*3,row))
        
        for i in just_colors:
            plt.subplot(row, col, pos)
            plt.hlines(0,0,5, color = i ,linestyles = 'solid', linewidth = 25)
            plt.axis('off')
            plt.text(0,0.04, i, fontsize = 20)
            pos = pos + 1

    plt.tight_layout()
    return plt.show()



# FUNCION 2
def visualizeME_and_describe_violinbox(dataframe, categ_var, numeric_var, palette= 'tab10', save= True):
    '''
    Function that allows to obtain a more complete graph by merging boxplot and violinplot together with a table of descriptive metrics
    It is high recommendable! to use this type of graph for a categoric variable with 20 unique values maximum.
    ### Parameters (5):
        * dataframe: `dataframe`  origin table
        * categ_var: `str` categoric variable
        * numeric_var:  `str` numeric variable
        * palette:  `str` by default 'tab10', but you can choose your palette
        * save:  `bool` by default True, the function save the plot and table generated
    '''
    # Generate ViolinBOX graph
    num_cat = len(list(dataframe[categ_var].unique()))
    plt.figure(figsize=(num_cat*1.5,10))
    sns.violinplot(x=categ_var, y=numeric_var, data=dataframe, palette= palette)
    ax = sns.boxplot(x=categ_var, y=numeric_var, data=dataframe,fliersize=0, color='white')
    ax.set_xticklabels(ax.get_xticklabels(), rotation=40, ha='right');
    titulo= numeric_var.upper() + '_vs_' + categ_var.upper()
    plt.title(titulo, fontsize=15);

    # Save graph
    if save == True:
        graph = 'visualizeME_Graphic_violinbox_' + titulo.lower() + '.png'
        plt.savefig(graph)

    # Metrics table
    cabeceras= ['Metrics',]
    fila1 = ['Upper limit',]
    fila2 = ['Q3',]
    fila3 = ['Median',]
    fila4 = ['Q1',]
    fila5 = ['Lower limit',]  
    iqr_ = iqr(dataframe[numeric_var], nan_policy='omit')
    d = [ fila1, fila2, fila3, fila4, fila5]
    for i in sorted(list(dataframe[categ_var].unique())):
        cabeceras.append(i)
        mediana = round(float(dataframe[dataframe[categ_var].isin([i])][[numeric_var]].median()), 2)
        fila3.append(mediana)
        q1 = round(np.nanpercentile(dataframe[dataframe[categ_var].isin([i])][[numeric_var]], 25), 2)
        fila4.append(q1)
        q3 = round(np.nanpercentile(dataframe[dataframe[categ_var].isin([i])][[numeric_var]], 75), 2)
        fila2.append(q3)
        th1 = round(q1 - iqr_*1.5, 2)
        fila5.append(th1)
        th2 = round(q3 + iqr_*1.5, 2)
        fila1.append(th2)
    table = pd.DataFrame(d, columns=cabeceras)
    table = table.set_index('Metrics')
    
    # Save table
    if save == True:
        name = 'visualizeME_table_violinbox_' + titulo.lower() + '.csv'
        table.to_csv(name, header=True)
    
    plt.show()
    display(table)



# FUNCION 3
def visualizeME_and_describe_barplot(dataframe, categ_var, numeric_var, palette='tab10', save = True):
    '''
    Function that allows to obtain a barplot with a table of descriptive metrics
    ### Parameters (5):
        * dataframe: `dataframe`  origin table
        * categ_var: `str` categoric variable
        * numeric_var:  `str` numeric variable
        * palette:  `str` by default 'tab10', but you can choose your palette
        * save:  `bool` by default True, the function save the plot and table generated
    '''
    # Graph
    num_cat = len(list(dataframe[categ_var].value_counts().index[::-1]))
    plt.figure(figsize=(num_cat*1.5,8))
    ax = sns.barplot(x= categ_var,y= numeric_var, data= dataframe, palette= palette, order= dataframe[categ_var].value_counts().index[::-1], ci=None)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=40, ha='right');
    titulo= numeric_var.upper() + ' vs. ' + categ_var.upper()
    plt.title(titulo, fontsize=15)
    
    # Save graph
    if save == True:
        path=os.path.join('visualizeME_Graphic_barplot_' + titulo.lower() + '.' + 'png')
        plt.savefig(path, format='png', dpi=300)
    
    # Metrics table
    eti_media= 'Mean ' + numeric_var
    eti_desv = 'Standard Deviation ' + numeric_var
    cabeceras= ['Metrics',]
    fila1 = ['Number of records',]
    fila2 = [eti_media,]
    fila3 = [eti_desv,]
    d = [fila1, fila2, fila3]
    for i in list(dataframe[categ_var].value_counts().index[::-1]):
        cabeceras.append(i)
        total = str(int(dataframe[dataframe[categ_var].isin([i])][[numeric_var]].count()))
        fila1.append(total)
        media = round(float(dataframe[dataframe[categ_var].isin([i])][[numeric_var]].mean()), 2)
        fila2.append(media)
        desv = round(float(dataframe[dataframe[categ_var].isin([i])][[numeric_var]].std()), 2)
        fila3.append(desv)
    table = pd.DataFrame(d, columns=cabeceras)
    table = table.set_index('Metrics')
    
    # Save table
    if save == True:
        name = 'visualizeME_table_barplot_' + titulo.lower() + '.csv'
        table.to_csv(name, header=True)
    
    plt.show()
    display(table)



# FUNCION 4
def visualizeME_FigureWords(dataframe, categ_var, shape= 'seahorse', cmap= 'tab10', contour= 'steelblue', back_color = 'white', height= 18, width = 20, save= True):
    '''
    Function that returns graph of words with different shapes, with the possibility to choose between 'dino', 'heart', 'star', 'seahorse' and 'hashtag'. I hope you like it!
    ### Parameters (9):
        * dataframe: `dataframe` origin table
        * categ_var: `str` categoric variable
        * shape: `str` by default is 'seahorse' shape, but you can choose from this list: 'seahorse', 'dino', 'heart', 'star' and 'hashtag'.
        * cmap: `str` by default is 'tab10', but you can choose your palette of Seaborn. If you want to know which palettes are available you can call visualizeME_colors_palettes() function
        * contour: `str` by default is 'steelblue', but you can choose your favourite color
        * back_color: `str` by default is 'white', but you can choose your background color
        * height: `int` by default is 18, but you can select your preference on height of the figure
        * width:`int` by default is 20, but you can select your preference on width of the figure
        * save: `bool` by default is True in order to save your graph, but if you prefer don't save it, just choose 'False'
    ### Return (1):
        * plt.show(): graph with your figure(by default will be seahorse)
    '''
    # Shape
    while shape not in ['dino', 'heart', 'star', 'seahorse', 'hashtag']:    
        shape = input('Try again, what shape do you want for your figure words graph?\n*Dino\n*Heart\n*Star\n*Seahorse: ').lower()
    if shape == 'seahorse':
        figure = 'data/viz/seahorse_visualizeME.jpg'
    elif shape == 'dino':
        figure = 'data/viz/dino_steg_visualizeME.jpg'
    elif shape == 'heart': 
        figure = 'data/viz/corazon_visualizeME.png'
    elif shape == 'star':
        figure = 'data/viz/estrella-silueta_visualizeME.png'  
    elif shape == 'hashtag':
        figure = 'data/viz/hashtag-silueta_visualizeME.png'
    
    # Words
    words = ' '.join(map(str, dataframe[categ_var]))
    custom_mask = np.array(Image.open(figure))
    wordcloud = WordCloud(background_color=back_color,
                      width=2500,
                      height=2000,
                      max_words=500, 
                      contour_width=0.1, 
                      contour_color= contour, 
                      colormap= cmap,
                      scale =5,mask=custom_mask).generate(words)
    
    plt.figure(1, figsize = (height, width))
    plt.imshow(wordcloud, interpolation='bilinear')
    plt.axis('off')

    # Save Graph
    if save == True:
        figure = figure.split('/')[1]
        figure = figure.split('.')[0]
        name = 'visualizeME_Graphic_' + figure + '.png'
        plt.savefig(name)
    
    return plt.show()



# FUNCION 5
def visualizeME_bagel_look_top(dataframe, categ_var, top=0, cmap = 'tab10', circle=True, save=True):
    '''
    Function to generate a bagel graphic where you can select the top categories you want to see, or everyone by default
    ### Parameters (6):
        * dataframe: `dataframe` origin table
        * categ_var: `str` categoric variable
        * top: `int` by default is 0, but you can choose look the top n categories and their weights.
        * cmap: `str` by default is 'tab10', but you can choose your palette of Seaborn. If you want to know which palettes are available you can call visualizeME_colors_palettes() function
        * circle: `bool` by default is True, in orders to seems like a bagel, but you can choose select a pie
        * save: `bool` by default is True in order to save your graph, but if you prefer don't save it, just choose 'False'
    '''
    data_valores= dataframe[categ_var].value_counts()
    # Filter by top categories
    if top != 0:
        data_valores_top= data_valores[(top):]
        p=0
        for i in data_valores_top:
            p=i+p
        data_valores_top=pd.Series([p],index=[f"Resto [{len(data_valores_top.index)}]"])
        data_valores_max= data_valores[:(top)]
        data_valores=pd.concat([data_valores_max, data_valores_top])
    
    # Generate graph
    plt.figure(figsize=(10,10))
    plt.pie(data_valores.values, labels=data_valores.index, textprops={"fontsize":15}, startangle = 60, autopct='%1.2f%%', frame=False, colors= sns.color_palette(cmap))
    p=plt.gcf()
    if circle is True:
        my_circle=plt.Circle((0,0), 0.4, color="w")
        p.gca().add_artist(my_circle)
    titulo= 'DISTRIBUCIÓN DE ' + categ_var.upper()
    plt.title(titulo, fontsize= 20)

    # Save graph
    if save == True:
        path=os.path.join('visualizeME_Graphic_baggel_' + titulo.lower() + '.png')
        plt.savefig(path, format='png', dpi=300)
    
    # Generate table
    values_bagel = pd.DataFrame(dataframe[categ_var].value_counts())
    new_list = []
    for i in list(dataframe[categ_var].value_counts()):
        sumat = sum(list(dataframe[categ_var].value_counts()))
        peso = i/sumat
        new_list.append(peso)
    porcentaj_nums = list(map(lambda x : x * 100, new_list))
    porcentaj_round3 = list(map(lambda x : round(x,2), porcentaj_nums))
    porcentaj = list(map(lambda x : str(x) + '%', porcentaj_round3))
    values_bagel['Pesos(%)'] = porcentaj

    # Save table
    if save == True:
        name = 'visualizeME_table_bagel_' + titulo.lower() + '.csv'
        values_bagel.to_csv(name, header=True)
    
    plt.show()
    return display(values_bagel)



#FUNCION 6
def visualizeME_and_describe_Spidey(dataframe, save= True):
    '''
    ## IMPORTANT: This function its just for SCALED dataframe (just numeric variables).
    ## If you don't have scaled your variables, please first do it!!!
    This function  generate a polar chart with your numeric 
    variables in order to compare their means.
    ### Parameters (2):
        * dataframe: `dataframe` origin table
        * save: `bool` by default is True in order to save your graph, but if you prefer don't save it, just choose 'False'
    '''
    spidey = pd.DataFrame(dataframe.mean(), columns=['Means'])

    categories=list(dataframe.columns)
    categories+=categories[:1]
    num =len(categories)

    # variables means
    value=list(dataframe.mean())
    value+=value[:1]

    loc_label = np.linspace(start=0, stop=2*np.pi, num= num)

    plt.figure(figsize=(10,10))
    ax = plt.subplot(polar=True)
    plt.plot(loc_label, value)
    plt.fill(loc_label, value, 'blue', alpha=0.1)
    # use thetagrids to place labels at the specified angles using degrees
    lines, labels = plt.thetagrids(np.degrees(loc_label), labels=categories)

    # Comienza radar chart arriba y hacia la derecha las variables
    ax.set_theta_offset(np.pi / 2)
    ax.set_theta_direction(-1)
    ax.set_rlabel_position(0)

    titulo= 'SPIDEY CHART TO COMPARE MEANS OF SCALED NUMERIC VARIABLES'
    plt.title(titulo, y=1.1, fontdict={'fontsize': 18})
    plt.legend(labels=['Mean'],loc=(1, 1))

    # Save graph
    if save == True:
        path=os.path.join('visualizeME_Graphic_' + titulo.lower() + '.png')
        plt.savefig(path, format='png', dpi=300)

    # Save Table
    if save == True:
        name = 'visualizeME_table_' + titulo.lower() + '.csv'
        spidey.to_csv(name, header=True)

    plt.show()
    return display(spidey)



#FUNCION 7
def visualizeME_c_matrix(y_true, 
                        y_pred, 
                        title='',
                        categories=[],
                        rotate=False,
                        cmap='',
                        cbar=True,
                        metrics=True,
                        save=True):
    '''
    This function plots a prettier confusion matrix with useful annotations 
       
    and adds a pd.DataFrame with the most common metrics used in classification.
        
    You can use this function with binary or multiclass classification.
    
    The figsize is calculated based on the number of categories.
        
    For both cases:
        * In the main diagonal you get:
            * counts over total of class
            * percentaje over total observations
        * In the rest of the cells you get:
            * counts
            * percentaje over total observations
        * If counts is zero you get an empty cell
           
    In case the function founds binary catergories in your data then a binary
    
    matrix is displayed with the TN, FN, FP, TP tags.
         
    ### Parameters:
    
        y_true -- `array_like`
            True labels.

        y_pred -- `array_like`
            Predictions to compare with true labels.

        title='' -- `str`
            Title to be displayed and used to save file.

        categories=[] -- `list(str)`
            List of names of the classes to be displayed instead of numeric values.

        rotate=False -- `bool`
            Applies a rotation on xticklabels in case they overlap.

        cmap='' -- `Matplotlib colormap name or object, or list of colors` 
            If not provided default is 'Blues'.

        cbar=True -- `bool`
            Whether to draw a colorbar.
            
        metrics=True -- `bool`
            Displays a pd.DataFrame with the most common metrics.

        save=False -- `bool` 
            Saves plot and metrics (if metrics=True) to disk. If title='' default is 'classifier'.
        
    ### Return:
    
        cfm -- `sklearn.metrics.confusion_matrix`
        
        metrics -- `pd.DataFrame`
    '''
    sns.set_theme(style='whitegrid')                
    # Generate confusion matrix
    cfm = confusion_matrix(y_true, y_pred)
    
    # Set tipe of classifier: binary or not binary
    if len(cfm)==2:
        base = len(cfm)
        bin_classifier = True
        annot_kws = {"size": 10*base}
        title_fontsize = 12*base
        axis_fontsize = 10*base
        labels_fontsize = 10*base
        plt.figure(figsize=(4.5*base, 4.5*base))
        
    else:
        base = len(cfm)
        bin_classifier = False    
        annot_kws = {"size": 1.5*base}
        title_fontsize = 3.5*base
        axis_fontsize = 2.5*base
        labels_fontsize = 1.5*base
        plt.figure(figsize=(1.75*base, 1.75*base))
    
    # Calculate auxiliar data   
    cfm_rowsum  = np.sum(cfm, axis=1, keepdims=True)
    
    if bin_classifier:
        cfm_percent = cfm / np.sum(cfm).astype(float)
    else:
        cfm_percent = cfm / cfm_rowsum.astype(float)
    
    # Build empty matrix for labels
    labels = np.zeros_like(cfm).astype(str)
    
    # Iterate labels to write correct annot
    nrows, ncols = cfm.shape
    
    for i in range(nrows):
        for j in range(ncols):
            count = cfm[i, j]
            percent = cfm_percent[i, j]
            
            if i == j:
                sum = cfm_rowsum[i]
                labels[i, j] = f'{count} / {int(sum)}\n{percent:.2%}'
            elif count == 0:
                labels[i, j] = ''
            else:
                labels[i, j] = f'{count}\n{percent:.2%}'
    
    if bin_classifier:
        names = ['TN', 'FP', 'FN', 'TP']
        labels = [name + '\n' + label for name, label in zip(names, labels.flatten())]
        labels = np.asarray(labels).reshape(2, 2)
    
    # Set color map
    if cmap == '':
        cmap = 'Blues'
    else:
        pass
        
    # Generate heatmap
    ax = sns.heatmap(cfm,
                     annot=labels,
                     annot_kws=annot_kws,
                     fmt='',
                     square=True,
                     cmap=cmap,
                     cbar=cbar)
        
    # Define categories position
    cat_position = [(i + .5) for i in range(len(labels))]
    
    # Add label rotation        
    if rotate:
        xdegree=50
        plt.yticks(rotation=0)
    else:
        xdegree=0
        plt.yticks(rotation=0)
    
    # Set title label
    if title == '':
        title = 'TRUE VS. PREDICTED'
    else:
        pass
    
    ax.set_title(f'{title.upper()}', fontsize=title_fontsize)
    
    # Set axis labels
    ax.set_xlabel('PREDICTED LABEL', fontsize=axis_fontsize)
    ax.xaxis.set_label_position('bottom')
    ax.set_ylabel('TRUE LABEL', fontsize=axis_fontsize)
    ax.yaxis.set_label_position('left')
    
    # Set tick labels
    # Define category names
    if categories != []:
        try:
            categories = [label for label in categories]
            ax.set_xticks(cat_position)
            ax.set_xticklabels(categories, fontsize=labels_fontsize, rotation=xdegree)
            ax.xaxis.tick_bottom()
            
            ax.set_yticklabels(categories, fontsize=labels_fontsize)
            ax.yaxis.tick_left()

        except ValueError:
            print('''Impossible to parse categories with number of classes. Ticklabels set to numeric.''')
            
    # Save plot
    if save:
        name = 'visualizeME_cf_matrix_' + title.lower() + '.png'
        path=os.path.join(name + '.' + 'png')
        plt.savefig(path, format='png', dpi=300)

    # Plot
    plt.show()
        
    # Calculate metrics
    if metrics:
        if bin_classifier:
            metrics_df = pd.DataFrame({title: [f'{accuracy_score(y_true, y_pred):.10f}',
                                            f'{precision_score(y_true, y_pred):.10f}',
                                            f'{recall_score(y_true, y_pred):.10f}',
                                            f'{f1_score(y_true, y_pred):.10f}',
                                            f'{roc_auc_score(y_true, y_pred):.10f}']},
                                   index=[['Accuracy: (TP + TN) / TOTAL',
                                           'Precision: TP / (TP + FP)',
                                           'Recall: TP / (TP + FN)',
                                           'F1: harmonic mean (accuracy, recall)',
                                           'ROC AUC']])
        else:
            report = classification_report(y_true, y_pred)
            report = [line.split(' ') for line in report.splitlines()]

            header = [x.upper() for x in report[0] if x!='']

            index = []
            values = []

            for row in report[1:-5]:
                row = [value for value in row if value!='']
                if row!=[]:
                    index.append(row[0].upper())
                    values.append(row[1:])

            index.append('ACCURACY')
            values.append(['-', '-'] + [x for x in report[-3] if x != ''][-2:])
            index.append('MACRO AVG.')
            values.append([x for x in report[-2] if x != ''][-4:])
            index.append('WEIGHTED AVG.')
            values.append([x for x in report[-1] if x != ''][-4:])

            metrics_df = pd.DataFrame(data=values, columns=header, index=index)

        # Plot metrics
        display(metrics_df)
    
    # Save metrics
    if save:
        name = 'visualizeME_cf_matrix_' + title.lower() + '.csv'
        metrics_df.to_csv(name, header=True)

    return cfm, metrics



# FUNCION 8
def visualizeME_scores_models(y_true,models,bin_multi_classifier= 'bin',pallete='tab10',save = True):
    ''' 
    This is a function that allows you to quickly identify the best metrics from your Machine Learning models whether is binary or non binary target
    ### Parameters(4):
        * y_true: -- `array_like` 
            True labels.
        * models: 
            here you should put a diccionary that contains for keys, the model names (as a str) and for values their predictions to compare with true labels.
             This will help to join all the metrics in one data frame
             EXAMPLE: {'Linear Regression':preds_multi,'KNN Classifier':preds_multi2}
        * pallete: `str`
            Seaborn color pallete we have selected before
        * bin_multi_classifier:`bool`
            True if it is a binomial model (0,1) and  False for a multicategory classification
        * save=False -- `bool` 
            Saves plot and metrics (if metrics=True) to disk. If title='' default is 'classifier'.
    '''
    # extracting values and keys from models dictionary:
    df = pd.DataFrame()
    keys_modelo = list(models.keys())
    values_modelo = list(models.values())   
    
    #If we have selected a binomial classifier then:
    if bin_multi_classifier == True:
        #Elaborating a Data Frame that will give you 5 metrics from you ML Model:
        for i,j in enumerate(models):
            metrics_df = pd.DataFrame({str(j): [float(f'{accuracy_score(y_true, values_modelo[i]):.10f}'),
                                            float(f'{precision_score(y_true, values_modelo[i]):.10f}'),
                                            float(f'{recall_score(y_true, values_modelo[i]):.10f}'),
                                            float(f'{f1_score(y_true, values_modelo[i]):.10f}'),
                                            float(f'{roc_auc_score(y_true, values_modelo[i]):.10f}')]},
                                index=[['Accuracy: (TP + TN) / TOTAL',
                                           'Precision: TP / (TP + FP)',
                                           'Recall: TP / (TP + FN)',
                                           'F1: harmonic mean (accuracy, recall)',
                                           'ROC AUC']])
            # Here we are joining all the dataframe metrics from all the models we have selected:
            if i == 0:
                df=metrics_df
            else:
                df = df.join(other=metrics_df)

        #FINDING METRICS
        for y in range(len(df.index)):
            #Finding the best Accuracy:
            if y == 0:
                for i,j in enumerate(df.iloc[y]):
                    if j == float(df.iloc[y].max()):
                        best_acc = df.columns[i]  

                # Setting a bar plot to see the best Accuracy:
                num=len(df.columns)
                plt.figure(figsize=(num*1.5,5))
                graf1= sns.barplot(x=df.columns,y=df.iloc[y],palette=pallete)
                graf1.set_xlabel('Metrics from Models',fontsize = 12)
                graf1.set_title('Accuracy',fontsize=15,color='blue')
                for i in graf1.patches:
                    graf1.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
                print('The best Accuracy value is '   + str(df.iloc[y].max())+' from Model: '+ str(best_acc))

                # Saving Bar plot
                if save == True:
                    path=os.path.join('visualizeME_Accuracy_barplot' + '.' + 'png')
                    plt.savefig(path, format='png', dpi=300)
                plt.show()
            
            #Finding the best Precision:
            elif y == 1:
                for i,j in enumerate(df.iloc[y]):
                    if j == float(df.iloc[y].max()):
                        best_prec = df.columns[i]    
            

                # Setting a bar plot to see the best Precision:
                num=len(df.columns)
                plt.figure(figsize=(num*1.5,5))
                graf2= sns.barplot(x=df.columns,y=df.iloc[y],palette=pallete)
                graf2.set_xlabel('Metrics from Models',fontsize = 12)
                graf2.set_title('Precision',fontsize=15,color='blue')
                for i in graf2.patches:
                    graf2.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
                print('The best Precision value is '   + str(df.iloc[y].max())+' from Model: '+ best_prec)
                
                
                # Saving Bar plot
                if save == True:
                    name = 'visualizeME_Precision_barplot' 
                    path=os.path.join(name + '.' + 'jpg')
                    plt.savefig(path, format='jpg', dpi=300)
                plt.show()

            #Finding the best Recall:
            elif y == 2:
                for i,j in enumerate(df.iloc[y]):
                    if j == float(df.iloc[y].max()):
                        best_rec = df.columns[i]    

                
                # Setting a bar plot to see the best Recall:
                num=len(df.columns)
                plt.figure(figsize=(num*1.5,5))
                graf3= sns.barplot(x=df.columns,y=df.iloc[y],palette=pallete)
                graf3.set_xlabel('Metrics from Models',fontsize = 12)
                graf3.set_title('Recall',fontsize=15,color='blue')
                for i in graf3.patches:
                    graf3.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
                print('The best Recall value is '   + str(df.iloc[y].max())+' from Model: '+ best_rec)
                
                # Saving Bar plot
                if save == True:
                    name = 'visualizeME_Recall_barplot'
                    path=os.path.join(name + '.' + 'jpg')
                    plt.savefig(path, format='jpg', dpi=300)
                plt.show()
            
            #Finding the best F1 score:
            elif y == 3:
                for i,j in enumerate(df.iloc[y]):
                    if j == float(df.iloc[y].max()):
                        best_f1 = df.columns[i]    
            

                # Setting a bar plot to see the best F1 Score:
                num=len(df.columns)
                plt.figure(figsize=(num*1.5,5))
                graf4= sns.barplot(x=df.columns,y=df.iloc[y],palette=pallete)
                graf4.set_xlabel('Metrics from Models',fontsize = 12)
                graf4.set_title('F1 Scores',fontsize=15,color='blue')
                for i in graf4.patches:
                    graf4.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
                print('The best F1 Score value is '   + str(df.iloc[y].max())+' from Model: '+ best_f1)
                
                # Saving Bar plot
                if save:
                    name = 'visualizeME_F1Score_barplot' + '.png'
                    path=os.path.join(name + '.' + 'png')
                    plt.savefig(path, format='png', dpi=300)
                plt.show()
            
            #Finding the best ROC:
            elif y == 4:
                for i,j in enumerate(df.iloc[y]):
                    if j == float(df.iloc[y].max()):
                        best_roc = df.columns[i]    
            
                # Setting a bar plot to see the best ROC AUC Score:
                num=len(df.columns)
                plt.figure(figsize=(num*1.5,5))
                graf5= sns.barplot(x=df.columns,y=df.iloc[y],palette= pallete)
                graf5.set_xlabel('Metrics from Models',fontsize = 12)
                graf5.set_title('ROC / AUC',fontsize=15,color='blue')
                for i in graf5.patches:
                    graf5.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
                print('The best ROC / AUC value is '   + str(df.iloc[y].max())+' from Model: '+ best_prec)
                
                # Saving Bar plot
                if save == True:
                    name = 'visualizeME_ROC_AUC_barplot' + '.png'
                    #path=os.path.join(name + '.' + 'png')
                    plt.savefig(name,format='png', dpi=300)
                plt.show()


    # bin_multi_classifier is False ->>>>> Multicategorical:
    else:
        for i,j in enumerate(models):
            report = classification_report(y_true,values_modelo[i])
            report = [line.split(' ') for line in report.splitlines()]

            header = [x.upper()+' : '+keys_modelo[i] for x in report[0] if x!='']

            index = []
            values = []

            for row in report[1:-5]:
                row = [value for value in row if value!='']
                if row!=[]:
                    index.append(row[0].upper())
                    values.append(row[1:])

            index.append('ACCURACY')
            values.append(['-', '-'] + [x for x in report[-3] if x != ''][-2:])
            index.append('MACRO AVG.')
            values.append([x for x in report[-2] if x != ''][-4:])
            index.append('WEIGHTED AVG.')
            values.append([x for x in report[-1] if x != ''][-4:])

            metrics_df = pd.DataFrame(data=values, columns=header, index=index)
            if i == 0:
                df=metrics_df
            elif i == 1:
                df = df.join(other=metrics_df)
            else:
                df = df.join(other=metrics_df,) 
        
        #Creating a Data Frame with the Accuracy metrics from all the models:
        acc_list = []
        
        for i in range(2,len(df.loc['ACCURACY']),4):
                acc_list.append(float(df.loc['ACCURACY'][i]))
        acc_dic={}
        for j,k in enumerate(models):
            acc_dic[k] = acc_list[j]
        
        acc = pd.DataFrame(data= acc_dic,index=keys_modelo)
    
        # Setting a BarPlot to show the Best model Accuracy:
        num=len(acc.index)
        plt.figure(figsize=(num*1.9,4))
        graf1= sns.barplot(data=acc,palette=pallete)
        graf1.set_xlabel('ML Models - Results',fontsize = 12)
        graf1.set_title('Accuracy',fontsize=15,color='blue')
        for i in graf1.patches:
            graf1.annotate(round(i.get_height(),3),(i.get_x() + i.get_width() / 2, i.get_height()),
                        ha='center',va='baseline',fontsize=12,color='black',
                            xytext=(0, 1),textcoords='offset points')
        if save:
            name = 'visualizeME_Accuracy_barplot' + '.png'
            path=os.path.join(name + '.' + 'png')
            plt.savefig(path, format='png', dpi=300)
        plt.show()
    
    # Saving Metrics
    if save == True:
        name = 'visualizeME_df_ML_metrics' + '.csv'
        df.to_csv(name, header=True)





#imports
    #resize_image()
from cv2 import imread , resize , IMREAD_GRAYSCALE
from os import remove , walk
from os.path import splitext
import zipfile
from pathlib import PurePath

    #reduce_image()
from cv2 import resize , INTER_AREA

    #process_color()
from numpy import floor

    #reduce_color_palette()
from numpy import reshape

import pandas as pd
import datetime as dt

# resize list of images to same sizes / dimensions
def resize_image(directory_path,
                pixels_1,
                pixels_2,
                bw = False,
                img_name_list= None,
                zip_file = False):
    '''   
    Resize images in .jpg format. Its purpose is to have all the image set levelled off 
    to use it in a convolutional neural net. Returns a list of the resized images.
    
    Keywords arguments:
    directory_path -- directory containing the set of original images
    pixels_1 -- value for the horizontal axis (pixels)
    pixels_2 -- value for the vertical axis (pixels)
    img_name_list -- in case there is a list of images names to use (default: None)
        
    '''
    
    #List to append the resized images
    re_image_list = []
    admited_formats_list = ['.bmp' , '.dib' , '.jpeg' , '.jpg' , '.jpe' , '.jp2',
                            '.png' , '.webp' , '.pbm' , '.pgm' , '.ppm' , '.pxm',
                            '.pnm' , '.sr' , '.ras' , '.tiff' , '.tif' , '.exr' ,
                            '.hdr' , '.pic']
    
    #If images in zip_file, extract and delete original zip
    if zip_file:
        
        with zipfile.ZipFile(directory_path , 'r') as z:
            
            for name in z.namelist():
                
                if (img_name_list != None):
                    z.extract(name, '.\images_unzip')
                else:
                    z.extract(name, '.\images_unzip')
                    
        #Delete zip file
        remove(directory_path)
        
        #Search paths in directories and subdirectories for all the images
        img_path_list = []
        for path , folders , files in walk('.\images_unzip'):
            
            #If not a predefined list of images, take all paths
            if img_name_list == None:
                
                for name in files: 
                    file_ext = splitext(name)[1]
                    
                    if file_ext.lower() in admited_formats_list:
                        img_path_list.append(PurePath(path, name))
                    else:
                        continue  
                    
            #If a predefined list of images, take only the ones in list
            if img_name_list != None:
                
                for name in files: 
                    
                    if name in img_name_list:
                        file_ext = splitext(name)[1]
                        
                        if file_ext.lower() in admited_formats_list:
                            img_path_list.append(PurePath(path, name))
                        else:
                            continue

        #Resize images and append to list
        print('Nº images:' ,
              len(img_path_list), '------- Final shape: %sx%s pixels' % (pixels_1 , pixels_2))
        
        for image_path in img_path_list:
            if bw:
                img = imread(str(image_path) , flags=IMREAD_GRAYSCALE)
            else:
                img = imread(str(image_path))
            image = resize(img, (pixels_1, pixels_2))
            re_image_list.append(image)     
        
    #If not zip_file
    if not zip_file:
        
        img_path_list = []
        for path , folders , files in walk('.\images_unzip'):
            
            #If not a predefined list of images, take all paths
            if img_name_list == None:
                
                for name in files: 
                    file_ext = splitext(name)[1]
                    
                    if file_ext.lower() in admited_formats_list:
                        img_path_list.append(PurePath(path, name))
                    else:
                        continue  
                    
            #If a predefined list of images, take only the ones in list
            if img_name_list != None:
                
                for name in files: 
                    
                    if name in img_name_list:
                        file_ext = splitext(name)[1]
                        
                        if file_ext.lower() in admited_formats_list:
                            img_path_list.append(PurePath(path, name))
                        else:
                            continue
                        
        #Resize images and append to list            
        print('Nº images:' , 
              len(img_name_list), '------- Final shape: %sx%s pixels' % (pixels_1 , pixels_2))
        
        for image_path in img_path_list:
            if bw:
                img = imread(str(image_path) , flags=IMREAD_GRAYSCALE)
            else:
                img = imread(str(image_path))
            image = resize(img, (pixels_1, pixels_2))
            re_image_list.append(image)
        
    return re_image_list


# standarize list of numbers into a format which makes possible to convert to float / integers
def standarize_numbers(numbers_list, 
                        type):
    '''
    Convert a list of numbers in string format to float format in case of 
    incompatibilities. Returns a list of float numbers (for "decimal" and "double") 
    or a list of integers (for "integer").
    
    Keywords arguments:
    numbers_list -- list of elements to convert
    type -- number type to convert. 3 values:
        - "decimal": decimal number denoted by "," instead of ".". Ej.: 6,67
        - "integer": integer number with thousand, millions, billions, etc. denoted by '.'. Ej.: 12.000.000
        - "double": number written in both previous forms. Ej.: 6.000.000,255 
    '''
    
    #Make transformations by 'type' argument value
    if type == 'decimal':
        float_list = list(map(lambda num: float(num.replace(',' , '.')) , numbers_list))
    elif type == 'integer':
        float_list = list(map(lambda num: int(num.replace('.' , '')) , numbers_list))
    elif type == 'double':
        float_list = list(map(lambda num: float(num.replace('.' , '').replace(',' , '.')) ,
                              numbers_list))
    else:
        print('Type not valid')

    return float_list


# transform str decimal coordenates into float decimal coordenates
def coordinates(list_lat,
                list_lng,
                tuple_format= False):
    '''
    Function to transform mixed coordinates (integers + cardinal point) to decimal 
    coordenates. Be careful, the two firts numbers will be taken as the integer part 
    of the decimal coordinate. Return 2 options: separated lists for latitude and 
    longitude or list of tuples representing geographic points.
    
    Keyword arguments:
    list_lat -- list of latitude values. Ej.: 400782N , 031213s 
    list_lng -- list of longitud values. Ej.: 239872e , 328769W (also accepts Spanish version "O")
    tuple_format -- If False, returns separated lists; if True, returns a list of tuples (default False)
    '''
    
    #List to append new latitude values and count to show position in case of error
    lat = []
    contador = 0
    
    #Loop to take all values in original latitude list
    for coor in list_lat:
        #Check if the original value is correctly written, if not, adds '0' before value to avoid errors
        if len(coor) < 7:
            coor = (7-len(coor))*'0' + coor
            
        #North values
        if (coor[-1] == 'N') or (coor[-1] == 'n'):
            valor = coor[0 : 2] + '.' + coor[2 : -1]
            lat.append(float(valor))
            #South values
        elif (coor[-1] =='S') or (coor[-1] =='s'):
            valor = coor[0 : 2] + '.' + coor[2 : -1]
            lat.append(-float(valor))
        #Anything else
        else:
            print('Latitudinal coordenate (%s) not valid in position (%s)' % (coor , contador))
            break
        
        contador += 1
        
    #List to append new longitude values and count to show position in case of error
    lng = []
    contador = 0
    #Loop to take all values in original longitude list
    for coor in list_lng:
        #Check if the original value is correctly written, if not, adds '0' before value to avoid errors
        if len(coor) < 7:
            coor = (7-len(coor))*'0' + coor
            
        #West values
        if (coor[-1] == 'W') or (coor[-1] == 'w') or (coor[-1] == 'O') or (coor[-1] == 'o'):
            valor = coor[0 : 2] + '.' + coor[2 : -1]
            lng.append(-float(valor))
        #East values
        elif (coor[-1] == 'E') or (coor[-1] == 'e'):
            valor = coor[0 : 2] + '.' + coor[2 : -1]
            lng.append(float(valor))
        else:
            print('Longitudinal coordenate (%s) not valid in position (%s)' % (coor , contador))
            break
        
        contador += 1
    
    #If taple_format, return list of coordenate tuples with zip(). If not, return separated lists
    if tuple_format:
        coor = list(zip(lat , lng))
        
        return coor
    
    else:
        
        return lat , lng


# reduce image size maintaining the original proportions
def reduce_img(image, height):
    '''
    Reduce the image dimensions, maintaining the ratio between original width 
    and height values. Returns the reduced image.
    
    Keyword arguments:
    image -- image to reduce dimensions
    height -- new height
    '''
    
    #Set dimensions ratio
    ratio = image.shape[0]/image.shape[1]
    #Calculate new width
    width = int(height/ratio)
     
    return resize(image, (width, height), interpolation=INTER_AREA)      


# process color image to use in the 'reduce_color_palette' function
def process_color(channel_value, bins):
    '''
    This function is used to map the values of RGB in a pixel that
    go from 0 to 255 to a simple version with X values that lead to
    a palette of Y colors (Xx Xx Xx = Y).
    Eg. X = 4, then y = 4x4x4 = 64 colors
    
    It is used in 'reduce_col_palette' function
    
    Keywords arguments:
    channel_value -- value of color intensity for each pixel in the BGR layers
    bins -- number of colours to reduce each layer (total colors = bins^3)
    '''
    
    ######################################
    if channel_value >= 255: 
        processed_value = 255
    else:
        preprocessed_value = floor((channel_value*bins)/255)
        processed_value = abs(int(preprocessed_value*(255/(bins-1))))
    
    return processed_value


# reduce the number of colours of a image
def reduce_col_palette(image, bins, info=False):
    '''
    This function iterate through every pixel of an image to map
    each rgb channel color value to a reduced palette.
    
    Keywords arguments:
    image -- image to reduce color palette
    bins -- number of colours to reduce each layer (total colors = bins^3). Used in 'process_color' function
    info -- if True, prints a message with the number of color of the new image
    '''
    
    # Capture image dimensions
    img = image.flatten()
       
    # Iterate the array to transform the value of the pixels
    for px in range(len(img)):
        
        if img[px] == 255: img[px] = 255
        else: img[px] = process_color(img[px], bins)
    
    # Restore image shape
    img = reshape(img, image.shape)
        
    # Inform user if 'info' is True about the total number of colors in new image
    if info:
        print(f'Palette reduced to {bins**3} colors.')
            
    return img


# calculate math expectation of a set of features
def math_expect(data, column_name_groups= None, column_name_success= None, n_top= 100):
    '''
    Function that calculates the mathematical expectation of a random variable ('column_name_groups').
    Definition: Mathematical expectation, also known as the expected value, is the the sum or integral of all possible
    values of a random variable, or any given function of it, multiplied by the respective probabilities of the values 
    of the variable. Return the DataFrame used as input containing a new column named 'math_expectation' 
    with the resulting values.
    
    Keyword arguments
    
    data (Pandas DataFrame type) -- is the given dataframe
    column_name_groups (string) -- name of column with the groups (random variable) to calculate the math expect
    column_name_success (string) -- name of column that sets the success measurement. ie: daily income or scoring 
    from movies reviews    
    n_top (integer) -- default 100, number of occurrencies within the top shortlist. ie: top 100, top 1000 or top 10
    '''
    len_data = len(data)

    #WE GROUP BY THE COLUMN WITH THE CHOSEN RANDOM VARIABLE
    groups = pd.DataFrame(data.groupby(column_name_groups)[column_name_groups].count())
    groups.rename(columns={column_name_groups:'groups'}, inplace=True)
    groups.reset_index(inplace=True)

    #WE ADD A COLUMN WITH THE DISTRIBUTION OF EACH GROUP IN %
    groups['% per group'] = (groups['groups'] / len_data) * 100

    #WE CREATE A COLUMN TO SEE THE EXPECTED NUMBER PER GROUP IN THE TOP LIST
    groups['expected_in_top'] = (groups['% per group'] * 10)

    #WE SORT BY THE SUCCESS VARIABLE UNTIL THE N_TOP FIGURE. IE: TOP 10, TOP 100, TOP 1000, ETC
    data.sort_values(by=column_name_success, ascending=False, inplace=True)
    top = data[:n_top]

    #WE GROUP BY AND CALCULATE THE ACTUAL NUMBER OF EACH GROUP VARIABLE IN THE TOP LIST
    top = top.groupby(column_name_groups)[column_name_groups].count()
    top.sort_values(ascending=False, inplace=True)
    top = pd.DataFrame(data=top)
    top.rename(columns={column_name_groups:'actual_top'}, inplace=True)
    top.reset_index(inplace=True)

    #WE MERGE THE TWO DATAFRAMES
    #WE REPLACE THE NAN VALUES FOR ZEROS
    join_group_top = groups.merge(top, how='left', on=column_name_groups)
    join_group_top['actual_top'] = join_group_top['actual_top'].fillna(0)

    #WE CREATE THE MATH_EXPECTATION COLUMN WITH ITS FORMULA AND SORT IT OUT BY THAT COLUMN
    join_group_top['math_expectation'] = ((join_group_top['actual_top'] - join_group_top['expected_in_top']) / join_group_top['expected_in_top'])*100
    join_group_top.sort_values(by='math_expectation', ascending=False, inplace=True)

    #WE CREATE A DICTIONARY WITH THE GROUP NAMES AND MATH_EXPECTATION VALUES
    dicc = {}
    for i, j in zip(join_group_top[column_name_groups], join_group_top['math_expectation']):
        dicc[i] = j

    #WE MAP THE RESULTS AND SORT BY MATH_EXPECTATION COLUMN
    data['math_expectation'] = data[column_name_groups].map(dicc) 
    data.sort_values(by='math_expectation', ascending=False, inplace=True)

    return data


# dataframe overview info
def overview(df):
    '''
    Function that summarizes data in a new dataframe from an originally given dataframe. It returns a new DataFrame
    with valuable information from the original DataFrame you are working on.
  
    Keyword arguments
    data (Pandas DataFrame type) -- the given dataframe
    '''
    #COLUMN NAMES
    cols = pd.DataFrame(df.columns.values,columns=['column names'])
    
    #COLUMN TYPES
    types = pd.DataFrame(df.dtypes.values, columns=["type of data"])

    #NAN TOTAL
    total_nan = df.isna().sum()
    total_nan = pd.DataFrame(total_nan.values, columns = ["nan total"])

    #NAN %
    percent_nan = round(df.isna().sum()*100/len(df),2)
    percent_nan_df = pd.DataFrame(percent_nan.values, columns = ["nan %"])

    #MISSINGS %
    percent_missing = round(df.isnull().sum()*100/len(df),2)
    percent_missing_df = pd.DataFrame(percent_missing.values, columns = ["missing %"])
 
    #UNIQUE VALUES
    unique = pd.DataFrame(df.nunique().values, columns = ["unique values"])
    
    #CARDINALITY
    percent_cardin = round(unique["unique values"]*100/len(df),2)
    percent_cardin_df = pd.DataFrame(percent_cardin.values, columns = ["cardin %"])

    #WE CONCAT THE DATAFRAMES
    concatenate = pd.concat([cols, types, total_nan, percent_nan_df, percent_missing_df, unique, percent_cardin_df], axis=1)
    concatenate.set_index('column names', drop=True, inplace=True)

    #WE ADD THE DESCRIBE() FUNCTION RESULTS AND CONCATENATE THEM TO THE PREVIOUS DATAFRAME
    df_describe = pd.DataFrame(df.describe(include= 'all'), columns=df.columns)
    df_concat = pd.concat([concatenate.T, df_describe])

    return df_concat


# removes outliers from a feature and related entries
def outlier_removal(data, column_name= None, p_max=None, p_min=None):
    '''
    Function that removes outliers from a given dataframe, specifying the column.
    min and max percentiles are passed as parameters to define the line between outliers and non-outliers.
    Returns a new DataFrame without the data below the p_min and above the p_max.
    
    Keyword arguments
    data (Pandas DataFrame type) -- is the given dataframe
    column_name (string) -- the chosen column to detect and remove outliers from
    p_max (integer) -- above this percentile we start removing outliers
    p_min (integer) -- below this percentile we start removing outliers
    '''
    q_max = p_max / 100
    q_min = p_min / 100
    q_max_column = data[column_name].quantile(q_max)
    q_min_column = data[column_name].quantile(q_min)

    data = data[(data[column_name] > q_min_column) & (data[column_name] <= q_max_column)]
    return data 


# transform columns to separated data values
def datetime_into_columns(df, column, weekday = False, hour_minutes = False, from_type = 'object'):
    """ 
    The function converts a column with a date from either int64 or object type
    into separate columns with day - month - year
    user can choose to add weekday - hour - minutes columns

    Keyword arguments
    df (Pandas DataFrame type)-- is the given dataframe
    column (string) -- the chosen column to create new columns from
    weekday (boolean) -- True if user wants new column with weekday value (default False)
    hour_minutes (boolean) -- True if user wants two new columns with hour and minutes values (default False)
    from_type (string) -- 'object' by default if original column type is object and
                        'int64' if original column type is int64

    return: the resulting dataframe with the new colum
    """

    if from_type == 'int64':
        column  = pd.to_datetime(df[column].astype(str))
    else:
        column  = pd.to_datetime(df[column])
        
    datetime = pd.DataFrame(column)

    datetime['day'] = column.dt.day
    datetime['month'] = column.dt.month
    datetime['year'] = column.dt.year

    if weekday == True:
        datetime['weekday'] = column.dt.weekday

    if hour_minutes == True:
        datetime['hour'] = column.dt.hour
        datetime['minutes'] = column.dt.minute

    df = pd.concat([df, datetime], axis = 1)
    df = df.loc[:,~df.columns.duplicated(keep='last')]
    
    return df


# treatment of missing data (NaNs) by column
def missing_data(df, columns = None, method = 'delete'):
    """
    The function allows user to decide whether they want to delete, convert to zeros, or replace by the 
    mean / mode / median all missing values in a specific list of columns of the given df. Returns the 
    resulting dataframe.

    Keyword arguments
    df (Pandas DataFrame type) -- is the given dataframe
    columns (list) -- the chosen column/s with missing values
    method (string) -- how to treat the NaN values
        - 'delete' drops the missing values (by default)
        - 'zero' fills missing values with a zero
        - 'mean' fills missing values with the mean
        - 'mode' fills missing values with the mode
        - 'median' fills missing values with the median
    """
 
    if method == 'delete':
        df.dropna(subset=columns, inplace= True)    
    elif method == 'zero':
        
        for column in columns:
            df[column] = df[column].fillna(0)
            
    elif method == 'mean':
        
        for column in columns:
            df[column] = df[column].fillna(df[column].mean())
            
    elif method == 'mode':
        
        for column in columns:
            df[column] = df[column].fillna(df[column].mode()[0])
            
    elif method == 'median':
        
        for column in columns:
            df[column] = df[column].fillna(df[column].median())
            
    return df



from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import RandomUnderSampler
from sklearn.linear_model import LinearRegression, Ridge, Lasso, ElasticNet, LogisticRegression
from sklearn.model_selection import train_test_split, GridSearchCV
from sklearn.preprocessing import StandardScaler, MinMaxScaler, PolynomialFeatures
from sklearn.metrics import accuracy_score, precision_score, recall_score, f1_score, roc_auc_score, mean_absolute_error, \
    mean_squared_error, mean_absolute_percentage_error
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
import pandas as pd
import numpy as np


def divide(dataset, target_column):
    """
    Divides the dataset in X and y.
    This function is called inside prepare_data function (see below) in order to prepare data.

    :param dataset:
    :param target_column:
    :return: X and y
    """
    X = dataset.drop(columns=[target_column])
    y = dataset[target_column]

    return X, y


def prepare_data(dataset, target_column, sampler_type=None, scaler_type='std', test_size=0.2):
    """
        Creates a list with several train-test sets to test with:

        1. Divides the dataset in X and y
        2. If scaler_type is under/smote/both, it samples the data. It will append to the list the generated train-test sets
        3. After that, it will scale all sets
        4. Returns A LIST OF LISTS with the following format:
            [
                [X_train, X_test, y_train, y_test] #set original
                [X_train, X_test, y_train, y_test] #Solo si se samplean los datos
            ]

        ¡IMPORTANT! The output of this function is used as input in all the model functions


        EXAMPLE:

        If we only want to use the original set, we can just use the first set of the array.
        Let's create a linear regression with library functions:
            sets = prepare_data(dataset, 'target')
            model, metrics = lin_reg(sets[0][0], sets[0][1], sets[0][2], sets[0][3])

        If sample is needed, this function will return more than one set. In that case, we will need to iterate over it:
            sets = prepare_data(dataset, 'target', sample_type='SMOTE')
            for i in range(len(sets)):
                model, metrics = lin_reg(sets[i][0], sets[i][1], sets[i][2], sets[i][3])


        :param pandas.__class__ dataset: Dataset we are working with
        :param str target_column: Target column name
        :param str sampler_type: Sampling to be applied. Accepts under, SMOTE, both and none
        :param str scaler_type: Scaler to be applied. Accepts std, minmax and None
        :param float test_size: Test size
        :return list: tuple list with X and y and sampled X and y
    """
    sets = []
    splited_sets = []
    scaled_sets = []

    # 1. Divide in X and y
    X, y = divide(dataset, target_column)
    sets.append((X, y))

    # 2. Sample if necessary
    if sampler_type == 'under':
        X_unders, y_unders = RandomUnderSampler(random_state=42).fit_resample(X, y)
        sets.append((X_unders, y_unders))

    elif sampler_type == 'SMOTE':
        X_overs, y_overs = SMOTE(random_state=42).fit_resample(X, y)
        sets.append((X_overs, y_overs))

    elif sampler_type == 'both':
        X_unders, y_unders = RandomUnderSampler(random_state=42).fit_resample(X, y)
        sets.append((X_unders, y_unders))
        X_overs, y_overs = SMOTE(random_state=42).fit_resample(X, y)
        sets.append((X_overs, y_overs))

    for set in sets:
        X_train, X_test, y_train, y_test = train_test_split(set[0], set[1], test_size=test_size,
                                                            random_state=42)
        splited_sets.append([X_train, X_test, y_train, y_test])

    # Optimize memory usage
    del sets

    for set in splited_sets:
        if scaler_type == "std":
            std_scale = StandardScaler().fit(set[0])  # fit(X_train)
            X_train_std = std_scale.transform(set[0])  # fit(X_train)
            X_test_std = std_scale.transform(set[1])  # fit(X_test)
            scaled_sets.append([X_train_std, X_test_std, set[2], set[3]])

        elif scaler_type == "minmax":
            minmax_scaler = MinMaxScaler().fit(set[0])  # fit(X_train)
            X_train_scl = minmax_scaler.transform(set[0])  # fit(X_train)
            X_test_scl = minmax_scaler.transform(set[1])  # fit(X_test)
            scaled_sets.append([X_train_scl, X_test_scl, set[2], set[3]])
        else:
            scaled_sets = splited_sets

    return scaled_sets


def model_scoring_regression(name, model, X_test, y_test):
    """
    Calculates model scoring dataframe for regression models
    This function is called inside regression model functions (see below) in order to generate metrics dataset.

    :param str name: Column name
    :param model: Trained model to get metrics of
    :param X_test: X_train
    :param y_test: y_test
    :return: metrics
    """
    name = f'{name.upper()} (test data)'
    preds = model.predict(X_test)

    metrics = pd.DataFrame({name: [f'{model.score(X_test, y_test):.10f}',
                                   f'{mean_absolute_error(y_test, preds):.10f}',
                                   f'{mean_absolute_percentage_error(y_test, preds):.10f}',
                                   f'{mean_squared_error(y_test, preds):.10f}',
                                   f'{np.sqrt(mean_squared_error(y_test, preds)):.10f}']},
                           index=[['Score (R2 coef.)', 'MAE', 'MAPE', 'MSE', 'RMSE']])

    return metrics


def lin_reg(X_train, X_test, y_train, y_test, regular_type=None):
    """
    Apply a lineal regression model with 3 polynomial levels
    :param X_train
    :param y_train
    :param X_test
    :param y_test
    :param regular_type: type of Regularization (ridge, lasso, elasticnet)
    :return trained model and metrics dataframe
    """

    model_fit = LinearRegression()

    model_fit.fit(X_train, y_train)

    metrics = model_scoring_regression('LinearRegression', model_fit, X_test, y_test)

    # Regularization if needed

    if regular_type == 'ridge':
        ridgeR = Ridge(alpha=10)
        ridgeR.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('LINREG-RIDGE', ridgeR, X_test, y_test)], axis=1)

    elif regular_type == 'lasso':
        lassoR = Lasso(alpha=1)
        lassoR.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('LINREG-LASSO', lassoR, X_test, y_test)], axis=1)

    elif regular_type == 'elasticnet':
        elastic_net = ElasticNet(alpha=1, l1_ratio=0.5)
        elastic_net.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('LINREG-ELASTICNET', elastic_net, X_test, y_test)], axis=1)

    elif regular_type == 'None':

        print("Regularization not necessary")

    return model_fit, metrics


def poly_reg(X_train, X_test, y_train, y_test, regular_type=None):
    """

    Apply a polynomial regression model with 2 polynomial levels

    :param X_train
    :param y_train
    :param X_test
    :param y_test
    :param regular_type: type of Regularization (ridge, lasso, elasticnet)
    :return trained model and metrics dataframe
    """

    poly_feats = PolynomialFeatures(degree=2)

    X_poly = poly_feats.fit_transform(X_train)

    model_fit = LinearRegression()

    model_fit.fit(X_poly, y_train)

    X_poly_test = poly_feats.transform(X_test)

    metrics = model_scoring_regression('PolynomialRegression', model_fit, X_poly_test, y_test)

    # Regularization if needed

    if regular_type == 'ridge':
        ridgeR = Ridge(alpha=10)
        ridgeR.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('POLREG-RIDGE', ridgeR, X_test, y_test)], axis=1)

    elif regular_type == 'lasso':
        lassoR = Lasso(alpha=1)
        lassoR.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('POLREG-LASSO', lassoR, X_test, y_test)], axis=1)

    elif regular_type == 'elasticnet':
        elastic_net = ElasticNet(alpha=1, l1_ratio=0.5)
        elastic_net.fit(X_train, y_train)
        metrics = pd.concat([metrics,
                             model_scoring_regression('POLREG-ELASTICNET', elastic_net, X_test, y_test)], axis=1)

    elif regular_type == 'None':

        print("Regularization not necessary")

    return model_fit, metrics


def model_scoring_classification(name, model, X_test, y_test, average="binary"):
    """
    Calculates model scoring for classification models
    This function is called inside model classification functions (see below) in order to generate metrics dataset.

    :param str name: Column name
    :param model: Trained model to get metrics of
    :param X_test: X_test
    :param y_test: y_test
    :param average: average param for precision, recall and score. If multiclass, use "weighted"
    :param multi_class: multi_class for roc_auc_score (ovr, ovo)
    :return: metrics dataframe
    """
    name = f'{name.upper()} (test data)'
    preds = model.predict(X_test)

    metrics = pd.DataFrame({name: [f'{accuracy_score(y_test, preds):.10f}',
                                   f'{precision_score(y_test, preds, average=average):.10f}',
                                   f'{recall_score(y_test, preds, average=average):.10f}',
                                   f'{f1_score(y_test, preds, average=average):.10f}']},
                           index=[['Accuracy (TP + TN/TT)', 'Precision (TP/TP + FP)', 'Recall (TP/TP + FN)',
                                   'F1 (har_mean Ac, Re)']])

    return metrics


def random_forest_classif(X_train, X_test, y_train, y_test):
    """
    Tests several parameters with a RandomForestClassifier model
    GridSearchCV is used with cv=10 and accuracy metric

    :param X_train: X_train
    :param X_test: X_test
    :param y_train: y_train
    :param y_test: y_test

    :returns: trained model, metrics dataframe and grid used
    """
    rand_forest = RandomForestClassifier()

    rand_forest_param = {
        'n_estimators': [100, 200, 400, 600, 800],
        'max_features': [2, 4, 6]
    }

    grid = GridSearchCV(rand_forest, rand_forest_param, cv=10, scoring='accuracy', n_jobs=-1, verbose=1)

    model_fit = grid.fit(X_train, y_train)

    metrics = model_scoring_classification('RandomForest', model_fit, X_test, y_test, average="weighted")

    return model_fit, metrics, grid


def SVC_model(X_train, X_test, y_train, y_test):
    """
    Tests several parameters with a SVC model
    GridSearchCV is used with cv=10 and accuracy metric

    :param X_train: X_train
    :param y_train: y_train
    :param X_test: X_test
    :param y_test: y_test

    :returns: trained model, metrics dataframe and grid used
    """

    svc = SVC()

    svc_param = {
        "C": np.arange(0.1, 1, 0.3),
        "gamma": ["scale", "auto"],
        "coef0": [-10., -1., 10],
        "kernel": ["linear", "poly", "rbf"],
    }

    grid = GridSearchCV(svc, svc_param, cv=10, scoring='accuracy', n_jobs=-1, verbose=1)

    model_fit = grid.fit(X_train, y_train)

    metrics = model_scoring_classification('SVC', model_fit, X_test, y_test, average="weighted")

    return model_fit, metrics, grid


def LogistRegress(X_train, X_test, y_train, y_test):
    """
    ests several parameters with a LogisticRegression model
    GridSearchCV is used with cv=10 and accuracy metric


    :param X_train: X_train
    :param y_train: y_train
    :param X_test: X_test
    :param y_test: y_test

    :return: fitted model, grid and metrics
    """

    log_reg = LogisticRegression()

    log_param = {
        'penalty': ['l2'],
        'C': np.arange(0.1, 4, 0.5),
        'solver': ['liblinear', 'newton-cg', 'lbfgs']
    }

    grid = GridSearchCV(log_reg, log_param, cv=10, scoring='accuracy', n_jobs=-1, verbose=1)

    model_fit = grid.fit(X_train, y_train)

    metrics = model_scoring_classification('LogisticRegression', model_fit, X_test, y_test, average="weighted")

    return model_fit, metrics, grid


# FUNCIÓN AUXILIAR PARA, DE UNA LISTA (EN PRINCIPIO, SIEMPRE UNA LISTA CON SCORE), PODER ESCOGER EL ELEMENTO MÁS ALTO
def change_score(element, p=0.90):
    """
    Function that iterates through the values parameter elements, converting them to positive (if necessary) and discarding those higher than the
    parameter value "p" (typically 0.90).
    It is useful to help the "knn" function (and the "corr_target" auxiliary function), with the aim of being able to identify the best score, and
    also discarding high values through the "p" parameter and thus avoiding OVERFITTING.
    :param element: List with elements (scores or correlations) to fit
    :param p: if the user prefer, percent to discard the high score or correlation. Dont could be higher than 1.
    :return: list with the changed values
    """
    if p >= 1:
        p = 0.90
    lista = []
    for i in element:
        if i < 0:
            i = - i
        if i >= 0.90:
            i = i - i
        lista.append(i)
    return lista


# FUNCIÓN PARA OBTENER LAS COLUMNAS CON MEJOR CORRELACIÓN CON EL TARGET
def corr_target(X, y, n_columns=1):
    """
    Function that identifies the two columns (max 4, if the user prefers) with the best correlation with respect to the target, of all Dataset.
    If the target is categorical, search the two columns with the best correlation between them.
    :param X: X_train from get the name of the best columns from correlation
    :param y: y_train
    :param n_columns: if the user prefer, numbers of columns to get from the correlation. Don't could be less than 1, higher than 4 or more than the
                      total of Dataset.
    :return: list with the name of the best columns from correlation. The first element always be the target, the second the column with higher
             correlation, the third the second one with higher correlation...
    """
    if n_columns <= 0:
        n_columns = 1

    df = X.join(y)

    lista_final = []

    if n_columns > len(df.columns):
        n_columns = len(df.columns)

    correlation_mat = df.corr()
    corr_pairs = correlation_mat.unstack()

    n = 0
    lista = []
    for i in corr_pairs.index:
        if i[0] == df.columns[-1]:
            lista.append(n)
        n = n + 1

    if len(lista) <= 0:  # <---Only in case the target is categorical
        correlation_mat = df.corr()
        corr_pairs = correlation_mat.unstack()
        lista = []
        lista = change_score(corr_pairs)
        mayor = max(lista)
        indice = lista.index(mayor)
        x = corr_pairs.index[indice]
        lista_final.append(df.columns[-1])
        lista_final.append(x[0])
        lista_final.append(x[1])
        return lista_final

    corr_pairs = corr_pairs[(lista[0]):(lista[-1] + 1)]

    lista_2 = []
    lista_2 = change_score(corr_pairs)

    lista_3 = lista_2.copy()
    lista_3.sort(reverse=True)

    if n_columns >= 4:
        mayor_4 = lista_3[3]
        indice_4 = lista_2.index(mayor_4)
        d = corr_pairs.index[indice_4]
        lista_final.append(d[1])

    if n_columns >= 3:
        mayor_3 = lista_3[2]
        indice_3 = lista_2.index(mayor_3)
        c = corr_pairs.index[indice_3]
        lista_final.append(c[1])

    if n_columns >= 2:
        mayor_2 = lista_3[1]
        indice_2 = lista_2.index(mayor_2)
        b = corr_pairs.index[indice_2]
        lista_final.append(b[1])

    if n_columns >= 1:
        mayor_1 = lista_3[0]
        indice_1 = lista_2.index(mayor_1)
        a = corr_pairs.index[indice_1]
        lista_final.append(a[1])
        lista_final.append(a[0])

    lista_final.reverse()

    return lista_final


# FUNCÓN PARA KNN
def knn(X_train, X_test, y_train, y_test, K=0, n_columns=2):
    """
    Function to implement a classification algorithm such as KNeighborsClassifier.
    :param X_train: X_train
    :param X_test: X_test
    :param y_train: y_train
    :param y_test: y_test
    :param K: if the user prefer, numbers of KNeighbors to fit the classifier
    :para n_columns: if the user prefer, number of columns to get the correlation to target. Only to "corr_target" auxiliary function
    :return: fitted model and metrics
    In case the Dataframe does not fit the classifier, returns None
    """
    try:
        if len(X_train.columns) < 2:
            print("Not enough columns")
            return None

        if len(X_train.columns) > 2:
            if n_columns < 2:
                n_columns = 2
            corr = corr_target(X_train, y_train, n_columns)
            colum1 = corr[1]
            colum2 = corr[2]

            X_train = pd.concat([X_train[colum1], X_train[colum2]], axis=1)
            X_test = pd.concat([X_test[colum1], X_test[colum2]], axis=1)

        scores_train = []
        scores_test = []
        k_range = range(1, 20)

        if K == 0:
            for i in k_range:
                knn_f = KNeighborsClassifier(n_neighbors=i)
                knn_f.fit(X_train, y_train)
                scores_train.append(knn_f.score(X_train, y_train))
                scores_test.append(knn_f.score(X_test, y_test))

            scores_change = change_score(scores_test)

            try:
                k = scores_test.index(max(scores_change)) + 1
            except:
                scores_test.reverse()
                k = scores_test.index(min(scores_test)) + 1
        else:
            if K < 0:
                K = -K
            k = K

        knn = KNeighborsClassifier(n_neighbors=k)
        model = knn.fit(X_train, y_train)

        metrics = model_scoring_classification('KNN', model, X_test, y_test, average="weighted")

        return model, metrics
    except ValueError:
        print("Dataframe does not fit classifier")
        return None



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
  name = 'Thorreznou',         
  version = '0.1.5',    
  author = 'The Bridge Data Science Team 1121',                   
  author_email = 'Thorreznou@gmail.com',
  description = '''Thorreznou.''',   
  long_description=long_description,
  long_description_content_type="text/markdown",
  url = 'https://github.com/erfederuiz/thorreznou',   
  classifiers=[
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
  ],
  package_dir={"": "src"},
  packages=setuptools.find_packages(where="src"),
  install_requires=[
    'pandas',
    'numpy',    
    'sklearn',
    'seaborn',
    'matplotlib',
    'scipy',
    'Ipython',
    'wordcloud',
    'pillow',
    'imblearn',
    'opencv-python',
    'pathlib'  
  ],
  python_requires=">=3.6",
)
